/* Copyright (C) 2005 Datapark corp. All rights reserved.

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA 
*/

#include "dps_common.h"
#include "dps_sea.h"
#include "dps_guesser.h"
#include "dps_log.h"
#include "dps_parsehtml.h"

#include <math.h>

#define f(x) (1.0 / (1.0 + exp(-1.0 * (x))))
#define EPS 0.0001
#define LOW_BORDER_EPS 0.000005
#define HI_BORDER_EPS (1.0 - 0.000005)

#define DPS_SEA_MAX_SENTN 100

const dpsunicode_t SentDelim[] = { '.', '!', '?', 0 };

static int SentCmp(const DPS_SENTENCE *s1, const DPS_SENTENCE *s2) {
  if (s1->w > s2->w) return -1;
  if (s1->w < s2->w) return 1;
  return 0;
}

int DpsSEAMake(DPS_AGENT *Indexer, DPS_DOCUMENT *Doc, DPS_DSTR *excerpt, DPS_CONV *uni_lc, 
	       const char *content_lang, size_t *indexed_size, size_t *indexed_limit, 
	       size_t max_word_len, size_t min_word_len, int crossec, int seasec
#ifdef HAVE_ASPELL
	       , int have_speller, AspellSpeller *speller
#endif
	       ) {
  DPS_SENTENCELIST List;
  DPS_MAPSTAT MapStat;
  DPS_TEXTITEM Item;
  dpsunicode_t *sentence, *lt;
  double *links;
  double di, Oi, delta, pas, pdiv, cur_div, newdi, dw;
  size_t l;
  int  it, u_it;
  
  bzero(&List, sizeof(List));
  sentence = DpsUniStrTok_R((dpsunicode_t*)excerpt->data, SentDelim, &lt);
  while(sentence && List.nitems < DPS_SEA_MAX_SENTN) {
    if (DpsUniLen(sentence) > 8) {
      if (List.nitems == List.mitems) {
	List.mitems += 16;
	List.Sent = (DPS_SENTENCE*)DpsRealloc(List.Sent, List.mitems * sizeof(DPS_SENTENCE));
	if (List.Sent == NULL) return DPS_ERROR;
      }
      List.Sent[List.nitems].sentence = sentence;
/*      List.Sent[List.nitems].w = 0.0001;*/
      bzero(&List.Sent[List.nitems].LangMap, sizeof(DPS_LANGMAP));
      for (l = 0; l <= DPS_LM_HASHMASK; l++) List.Sent[List.nitems].LangMap.memb[l].index = l;
      DpsBuildLangMap(&List.Sent[List.nitems].LangMap, (char*)sentence, DpsUniLen(sentence) * sizeof(dpsunicode_t), 0, 0);
      DpsPrepareLangMap(&List.Sent[List.nitems].LangMap);
      List.nitems++;
    }
    sentence = DpsUniStrTok_R(NULL, SentDelim, &lt);
  }
  DpsLog(Indexer, DPS_LOG_DEBUG, "SEA sentences: %d", List.nitems);
  if (List.nitems < 6) {
    DPS_FREE(List.Sent); return DPS_OK; 
  }

  links = (double*)DpsMalloc(sizeof(double) * List.nitems * List.nitems);
  if (links != NULL) {
    register size_t i, j;
    for (i = 0; i < List.nitems; i++) {
      List.Sent[i].w = 1.0 / List.nitems;
      for (j = 0; j < List.nitems; j++) {
	if (i == j) { links[i * List.nitems + j] = 0.0; continue; }
	MapStat.map = &List.Sent[j].LangMap;
	DpsCheckLangMap(&List.Sent[j].LangMap, &List.Sent[i].LangMap, &MapStat, DPS_LM_TOPCNT + 2);
	links[i * List.nitems + j] = ((double)MapStat.hits) / (MapStat.miss + 1) / DPS_LM_TOPCNT;
/*	DpsLog(Indexer, DPS_LOG_INFO, "Link %u->%u: %f [hits:%d miss:%d]", i, j, links[i * List.nitems + j], MapStat.hits, MapStat.miss);*/
      }
    }
    for (l = 0; l < List.nitems; l++) {
      Oi = 0.0;
      for (i = 0; i < List.nitems; i++) Oi += List.Sent[l].w * links[l * List.nitems + i];
      Oi = f(Oi);
      if (Oi < LOW_BORDER_EPS) Oi = LOW_BORDER_EPS;
      else if (Oi > HI_BORDER_EPS) Oi = HI_BORDER_EPS;
      List.Sent[l].w = Oi;
    }
    for (it = 0; it < Indexer->Flags.PopRankNeoIterations; it++) {

      for (l = 0; l < List.nitems; l++) {
	newdi = 0.0;
	for (i = 0; i < List.nitems; i++) newdi += List.Sent[i].w * links[i * List.nitems + l];
	if (newdi < LOW_BORDER_EPS) newdi = LOW_BORDER_EPS;
	else if (newdi > HI_BORDER_EPS) newdi = HI_BORDER_EPS;
	/*List.Sent[l].w = */di = newdi;

	Oi = 0.0;
	for (i = 0; i < List.nitems; i++) Oi += List.Sent[l].w * links[l * List.nitems + i];
	Oi = f(Oi);
	if (Oi < LOW_BORDER_EPS) Oi = LOW_BORDER_EPS;
	else if (Oi > HI_BORDER_EPS) Oi = HI_BORDER_EPS;
	List.Sent[l].w = Oi;

	pas = -1.07;
	pdiv = cur_div = fabs(di - Oi);
	u_it = ( cur_div > EPS );

/*	if (!u_it) break;*/

	delta = pas * (di - Oi) * di * (1.0 - di);
	dw = List.Sent[l].w * delta;

	if (fabs(delta) > 0.0) {
	    for (j = 0; j < List.nitems; j++) links[j * List.nitems + l] += dw;
	} else {
	  break;
	}

	newdi = 0.0;
	for (i = 0; i < List.nitems; i++) newdi += List.Sent[i].w * links[i * List.nitems + l];
	if (newdi < LOW_BORDER_EPS) newdi = LOW_BORDER_EPS;
	else if (newdi > HI_BORDER_EPS) newdi = HI_BORDER_EPS;
	/*List.Sent[l].w =*/ di = newdi;

	Oi = 0.0;
	for (i = 0; i < List.nitems; i++) Oi += List.Sent[l].w * links[l * List.nitems + i];
	Oi = f(Oi);
	if (Oi < LOW_BORDER_EPS) Oi = LOW_BORDER_EPS;
	else if (Oi > HI_BORDER_EPS) Oi = HI_BORDER_EPS;
	List.Sent[l].w = Oi;

	cur_div = fabs(di - Oi);
/*
	if ((cur_div > pdiv) && ((cur_div - pdiv) > EPS)) {
	  pas *= 0.73;
	} else if (fabs(delta) < 0.1) {
	  if (fabs(cur_div - pdiv) < 0.1 * pdiv) {
	    pas *= 9.99;
	  } else if (fabs(cur_div - pdiv) < 0.5 * pdiv) {
	    pas *= 2.11;
	  }
	} else if (fabs(delta) > 1.0) pas *= 0.95;
*/
	DpsLog(Indexer, DPS_LOG_DEBUG, "%d:%02d|%12.9f->%12.9f|di:%11.9f|Oi:%11.9f|delta:%12.9f|pas:%11.9f", 
	       l, it, pdiv, cur_div,  di, Oi, delta, pas);

/*	u_it = ( (pdiv = cur_div) > EPS );*/

      }
    }

    DpsSort(List.Sent, List.nitems, sizeof(DPS_SENTENCE), (qsort_cmp)SentCmp);
/*
    DpsUniPrint("Sent.0", List.Sent[0].sentence);
    fprintf(stderr, "\tw:%f\n", List.Sent[0].w);
    if (List.nitems > 1) { 
      DpsUniPrint("Sent.1", List.Sent[1].sentence);
      fprintf(stderr, "\tw:%f\n", List.Sent[1].w);
    }
    if (List.nitems > 2) { 
      DpsUniPrint("Sent.2", List.Sent[2].sentence);
      fprintf(stderr, "\tw:%f\n", List.Sent[2].w);
    }
*/
    Item.section = seasec;
    Item.href = NULL;
    Item.section_name = "sea";
    for (i = 0; i < 3; i++) {
      DpsPrepareItem(Indexer, Doc, &Item, uni_lc, List.Sent[i].sentence, List.Sent[i].sentence, content_lang, indexed_size, indexed_limit,
		     max_word_len, min_word_len, crossec
#ifdef HAVE_ASPELL
		     , have_speller, speller
#endif
		     );
    }
    DPS_FREE(links);
  }
  DPS_FREE(List.Sent);

  return DPS_OK;
}
